// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_CAST_HPP
#define FF_ST_CAST_HPP

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits, uint_fast8_t exp_bits2, uint_fast8_t frac_bits2>
inline void ff_st_cast(flexfloat_st<exp_bits2,frac_bits2> *obj, const flexfloat_st<exp_bits,frac_bits> *source) {
    flexfloat_desc_t desc;
    desc.exp_bits = exp_bits2;
    desc.frac_bits = frac_bits2;
    ff_cast(obj->x, source->x, desc);
    ff_cast(obj->y, source->y, desc);
    ff_cast(obj->z, source->z, desc);
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline float ff_st_get_float(const flexfloat_st<exp_bits,frac_bits> *obj) {
    float a, b, c;
    a = ff_get_float(obj->x);
    b = ff_get_float(obj->y);
    c = ff_get_float(obj->z);
    return (float)((a+b+c)/3.);
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline double ff_st_get_double(const flexfloat_st<exp_bits,frac_bits> *obj) {
    double a, b, c;
    a = ff_get_double(obj->x);
    b = ff_get_double(obj->y);
    c = ff_get_double(obj->z);
    return (double)((a+b+c)/3.);
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline long double ff_st_get_longdouble(const flexfloat_st<exp_bits,frac_bits> *obj) {
    long double a, b, c;
    a = ff_get_longdouble(obj->x);
    b = ff_get_longdouble(obj->y);
    c = ff_get_longdouble(obj->z);
    return (long double)((a+b+c)/3.);
}

#ifdef FF_ST_FLOAT128
template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline __float128 ff_st_get_longdouble(const flexfloat_st<exp_bits,frac_bits> *obj) {
    __float128 a, b, c;
    a = ff_get_float128(obj->x);
    b = ff_get_float128(obj->y);
    c = ff_get_float128(obj->z);
    return (__float128)((a+b+c)/3.);
}
#endif

#endif