// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_CONVERT
#define FF_ST_CONVERT

inline void data_st(const double &e, const int &error_type)
{
    if (error_type)
    {
        x += (e * rand() / (double)RAND_MAX);
        y = -((-y) - (-(e * rand() / (double)RAND_MAX)));
    }
    else
    {
        x *= ((double)1.0 + (e * rand() / (double)RAND_MAX));
        y = -(y * (-((double)1.0 - (e * rand() / (double)RAND_MAX))));
    }
    accuracy = DIGIT_NOT_COMPUTED;
}

inline void data_st()
{
    double aux = 1.0;
    if (x == y)
    {
        x *= (aux + 1.e-20);
        y = -(-y * (aux - 1.e-20));
    }
    accuracy = DIGIT_NOT_COMPUTED;
}

template <uint_fast8_t exp_bits2, uint_fast8_t frac_bits2>
operator flexfloat_st<exp_bits2, frac_bits2>() const
{
    flexfloat_st<exp_bits2, frac_bits2> res;
    res.x = x;
    res.y = y;
    res.z = z;
    return (res);
}

#ifdef FF_ST_FLOAT128
operator __float128()
{
    return ((__float128)(x / 3. + y / 3. + z / 3.));
}
#endif

operator double()
{
    return ((double)(x / 3. + y / 3. + z / 3.));
}

operator float()
{
    return ((float)(x / 3. + y / 3. + z / 3.));
}

#endif