// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_EQ_HPP
#define FF_ST_EQ_HPP


friend inline int operator==(const flexfloat_st &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b.x.getValue();
    else
        res.x = -((-a.x.getValue()) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b.y.getValue();
        res.z = -((-a.z.getValue()) - (-b.z.getValue()));
    }
    else
    {
        res.z = a.z.getValue() - b.z.getValue();
        res.y = -((-a.y.getValue()) - (-b.y.getValue()));
    }

    return res.computedzero();
}

#ifdef FF_ST_FLOAT128
friend inline int operator==(const flexfloat_st &a, const __float128 &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}
#endif

friend inline int operator==(const flexfloat_st &a, const long double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}

friend inline int operator==(const flexfloat_st &a, const double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}

friend inline int operator==(const flexfloat_st &a, const float &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}

friend inline int operator==(const flexfloat_st &a, const int &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}

#ifdef FF_ST_HALF_EMULATION
friend inline int operator==(const flexfloat_st &a, const half &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res.computedzero();
}
#endif

#ifdef FF_ST_FLOAT128
friend inline int operator==(const __float128 &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}
#endif

friend inline int operator==(const long double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}

friend inline int operator==(const double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}


friend inline int operator==(const float &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}

friend inline int operator==(const int &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}

#ifdef FF_ST_HALF_EMULATION
friend inline int operator==(const half &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res.computedzero();
}
#endif


#ifdef FF_ST_FLOAT128
friend inline int equal(const __float128 &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}
#endif

friend inline int equal(const long double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}

friend inline int equal(const double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}

friend inline int equal(const float &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}

friend inline int equal(const int &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}

#ifdef FF_ST_HALF_EMULATION
friend inline int equal(const half &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = b.x.getValue() - a;
    else
        res.x = -((-b.x.getValue()) - (-a));

    if (RANDOM)
    {
        res.y = b.y.getValue() - a;
        res.z = -((-b.z.getValue()) - (-a));
    }
    else
    {
        res.z = b.z.getValue() - a;
        res.y = -((-b.y.getValue()) - (-a));
    }

    return res < pow(10, b.getaccuracy());
}
#endif


#ifdef FF_ST_FLOAT128
friend inline int equal(const flexfloat_st &a, const __float128 &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}
#endif

friend inline int equal(const flexfloat_st &a, const long double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}

friend inline int equal(const flexfloat_st &a, const double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}

friend inline int equal(const flexfloat_st &a, const float &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}

friend inline int equal(const flexfloat_st &a, const int &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}

#ifdef FF_ST_HALF_EMULATION
friend inline int equal(const flexfloat_st &a, const half &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    return res < pow(10, a.getaccuracy());
}
#endif

#endif