// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_FMA_HPP
#define FF_ST_FMA_HPP


// INLINE void ff_fma(flexfloat_t *dest, const flexfloat_t *a, const flexfloat_t *b, const flexfloat_t *c) {
//     assert((dest->desc.exp_bits == a->desc.exp_bits) && (dest->desc.frac_bits == a->desc.frac_bits) &&
//            (a->desc.exp_bits == b->desc.exp_bits) && (a->desc.frac_bits == b->desc.frac_bits) &&
//            (b->desc.exp_bits == c->desc.exp_bits) && (b->desc.frac_bits == c->desc.frac_bits));
//     #ifdef FLEXFLOAT_ROUNDING
//     // Change the rounding mode according to the error direction if we need to do manual rounding for RNE
//     int mode = fegetround();
//     bool eff_sub = flexfloat_sign(a) ^ flexfloat_sign(b) ^ flexfloat_sign(c);
//     if (a->desc.frac_bits < NUM_BITS_FRAC && mode == FE_TONEAREST) {
//         if (!eff_sub) { // in this case, we need to round away from zero
//             fexcept_t flags;
//             fegetexceptflag(&flags, FE_ALL_EXCEPT); // get accrued flags to not tarnish them here
//             double try = fma(a->value, b->value, c->value);
//             (try >= 0) ? fesetround(FE_UPWARD) : fesetround(FE_DOWNWARD);
//             fesetexceptflag(&flags, FE_ALL_EXCEPT); // restore flags here
//         } else {
//             fesetround(FE_TOWARDZERO); // just truncate
//         }
//     }
//     #endif
//     dest->value = fma(a->value, b->value, c->value); // finally the actual operation
//     #ifdef FLEXFLOAT_TRACKING
//     dest->exact_value = fma(a->exact_value, b->exact_value, c->exact_value);
//     if(dest->tracking_fn) (dest->tracking_fn)(dest, dest->tracking_arg);
//     #endif
//     #ifdef FLEXFLOAT_ROUNDING
//     if (a->desc.frac_bits < NUM_BITS_FRAC && mode == FE_TONEAREST)
//         fesetround(FE_TONEAREST); // restore rounding
//     #endif
//     flexfloat_sanitize(dest);
//     #ifdef FLEXFLOAT_STATS
//     if(StatsEnabled) getOpStats(dest->desc)->fma += 1;
//     #endif
// }

friend flexfloat_st<exp_bits, frac_bits> fma(const flexfloat_st<exp_bits, frac_bits> &a, const flexfloat_st<exp_bits, frac_bits> &b, const flexfloat_st<exp_bits, frac_bits> &c)
{
    flexfloat_st<exp_bits, frac_bits> res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    // ff_fma(flexfloat_t * dest, const flexfloat_t *a, const flexfloat_t *b, const flexfloat_t *c);

    if (RANDOM)
    {
        res.x = fma(a.x.getValue(), b.x.getValue(), c.x.getValue());
    }
    else 
    {   
        res.x = -fma(a.x.getValue(), -b.x.getValue(), -c.x.getValue());
    }

    if (RANDOM)
    {
        res.y = fma(a.y.getValue(), b.y.getValue(), c.y.getValue());
        res.z = -fma(a.z.getValue(), -b.z.getValue(), -c.z.getValue());
    }
    else
    {
        res.y = -fma(a.y.getValue(), -b.y.getValue(), -c.y.getValue());
        res.z = fma(a.z.getValue(), b.z.getValue(), c.z.getValue());
    }

    if (_ff_st_mul_tag)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
        {
            if (b.accuracy == DIGIT_NOT_COMPUTED)
                b.approx_digit();
            if (b.accuracy == 0)
                instability(&_ff_st_mul_count);
        }
    }

    if (_ff_st_cancel_tag)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED || a.accuracy == RELIABLE_RESULT)
            a.nb_significant_digit();
        if (b.accuracy == DIGIT_NOT_COMPUTED || b.accuracy == RELIABLE_RESULT)
            b.nb_significant_digit();
        if (c.accuracy == DIGIT_NOT_COMPUTED || c.accuracy == RELIABLE_RESULT)
            c.nb_significant_digit();
        if ((a.accuracy < b.accuracy ? (a.accuracy < c.accuracy ? a.accuracy : c.accuracy) : (b.accuracy < c.accuracy ? b.accuracy : c.accuracy)) >=
            res.nb_significant_digit() + _ff_st_cancel_value)
            _ff_st_cancel_count++;
    }

    return res;
}

#endif