// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_GE_HPP
#define FF_ST_GE_HPP


friend inline int operator>=(const flexfloat_st &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b.x.getValue();
    else
        res.x = -((-a.x.getValue()) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b.y.getValue();
        res.z = -((-a.z.getValue()) - (-b.z.getValue()));
    }
    else
    {
        res.z = a.z.getValue() - b.z.getValue();
        res.y = -((-a.y.getValue()) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b.x + b.y + b.z));
}


#ifdef FF_ST_FLOAT128
friend inline int operator>=(const __float128 &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}
#endif

friend inline int operator>=(const long double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}

friend inline int operator>=(const double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}

friend inline int operator>=(const float &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}

friend inline int operator>=(const int &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}

#ifdef FF_ST_HALF_EMULATION
friend inline int operator>=(const half &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a - b.x.getValue();
    else
        res.x = -((-a) - (-b.x.getValue()));

    if (RANDOM)
    {
        res.y = a - b.y.getValue();
        res.z = -((-a) - (-b.z.getValue()));
    }
    else
    {
        res.z = a - b.z.getValue();
        res.y = -((-a) - (-b.y.getValue()));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a + a + a) >= (b.x + b.y + b.z));
}
#endif


#ifdef FF_ST_FLOAT128
friend inline int operator>=(const flexfloat_st &a, const __float128 &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}
#endif

friend inline int operator>=(const flexfloat_st &a, const long double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}

friend inline int operator>=(const flexfloat_st &a, const double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}

friend inline int operator>=(const flexfloat_st &a, const float &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}

friend inline int operator>=(const flexfloat_st &a, const int &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}

#ifdef FF_ST_HALF_EMULATION
friend inline int operator>=(const flexfloat_st &a, const half &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
        res.x = a.x.getValue() - b;
    else
        res.x = -((-a.x.getValue()) - (-b));

    if (RANDOM)
    {
        res.y = a.y.getValue() - b;
        res.z = -((-a.z.getValue()) - (-b));
    }
    else
    {
        res.z = a.z.getValue() - b;
        res.y = -((-a.y.getValue()) - (-b));
    }

    int r = res.numericalnoise();
    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    return r || ((a.x + a.y + a.z) >= (b + b + b));
}
#endif
#endif