// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_MUL_HPP
#define FF_ST_MUL_HPP



friend inline flexfloat_st &operator*=(flexfloat_st &a, const flexfloat_st &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (_ff_st_mul_tag)
    {
        if (b.accuracy == DIGIT_NOT_COMPUTED)
            b.approx_digit();
        if (b.accuracy == 0)
        {
            if (a.accuracy == DIGIT_NOT_COMPUTED)
                a.approx_digit();
            if (a.accuracy == 0){
                instability(&_ff_st_mul_count);
            }
        }
    }

    if (RANDOM)
    {
        a.x = a.x * b.x;
    }
    else
        a.x = -(a.x * -b.x);

    if (RANDOM)
    {
        a.y = a.y * b.y;
        a.z = -(a.z * -b.z);
    }
    else
    {
        a.y = -(a.y * -b.y);
        a.z = a.z * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}

#ifdef FF_ST_FLOAT128
friend inline flexfloat_st &operator*=(flexfloat_st &a, const __float128 &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}
#endif

friend inline flexfloat_st &operator*=(flexfloat_st &a, const long double &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}

friend inline flexfloat_st &operator*=(flexfloat_st &a, const double &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}

friend inline flexfloat_st &operator*=(flexfloat_st &a, const float &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}

friend inline flexfloat_st &operator*=(flexfloat_st &a, const int &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}

#ifdef FF_ST_HALF_EMULATION
friend inline flexfloat_st &operator*=(flexfloat_st &a, const half &b)
{

    a.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        a.x = a.x * b;
    }
    else
        a.x = -(a.x * -b);

    if (RANDOM)
    {
        a.y = a.y * b;
        a.z = -(a.z * -b);
    }
    else
    {
        a.y = -(a.y * -b);
        a.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf(a.x) or std::isinf(a.y) or std::isinf(a.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    #endif

    return a;
}
#endif

friend inline flexfloat_st operator*(const flexfloat_st &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (_ff_st_mul_tag)
    {
        if (b.accuracy == DIGIT_NOT_COMPUTED)
            b.approx_digit();
        if (b.accuracy == 0)
        {
            if (a.accuracy == DIGIT_NOT_COMPUTED)
                a.approx_digit();
            if (a.accuracy == 0){
                std::cout << "instab mult\n";
                instability(&_ff_st_mul_count);
            }
        }
    }

    if (RANDOM)
    {
        res.x = a.x * b.x;
    }
    else
        res.x = -(a.x * -b.x);

    if (RANDOM)
    {
        res.y = a.y * b.y;
        res.z = -(a.z * -b.z);
    }
    else
    {
        res.y = -(a.y * -b.y);
        res.z = a.z * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif

    return res;
}

#ifdef FF_ST_FLOAT128
friend inline flexfloat_st operator*(const __float128 &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif

    return res;
}
#endif

friend inline flexfloat_st operator*(const long double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif

    return res;
}

friend inline flexfloat_st operator*(const double &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

friend inline flexfloat_st operator*(const float &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

friend inline flexfloat_st operator*(const int &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend inline flexfloat_st operator*(const half &a, const flexfloat_st &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a * b.x;
    }
    else
        res.x = -(a * -b.x);

    if (RANDOM)
    {
        res.y = a * b.y;
        res.z = -(a * -b.z);
    }
    else
    {
        res.y = -(a * -b.y);
        res.z = a * b.z;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a!=0) && (b.x!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}
#endif

#ifdef FF_ST_FLOAT128
friend inline flexfloat_st operator*(const flexfloat_st &a, const __float128 &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}
#endif

friend inline flexfloat_st operator*(const flexfloat_st &a, const long double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

friend inline flexfloat_st operator*(const flexfloat_st &a, const double &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

friend inline flexfloat_st operator*(const flexfloat_st &a, const float &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

friend inline flexfloat_st operator*(const flexfloat_st &a, const int &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend inline flexfloat_st operator*(const flexfloat_st &a, const half &b)
{
    flexfloat_st res;

    res.accuracy = DIGIT_NOT_COMPUTED;

    if (RANDOM)
    {
        res.x = a.x * b;
    }
    else
        res.x = -(a.x * -b);

    if (RANDOM)
    {
        res.y = a.y * b;
        res.z = -(a.z * -b);
    }
    else
    {
        res.y = -(a.y * -b);
        res.z = a.z * b;
    }

    #ifdef FF_ST_HALF_EMULATION
    if(_ff_st_half_overflow_tag){
        if (std::isinf((double)res.x) or std::isinf((double)res.y) or std::isinf((double)res.z)){
            instability(&_ff_st_half_overflow_count);
        }
    }
    if(_ff_st_half_underflow_tag){
        if ( ( (res.x==0) || (res.y==0) || (res.z==0) ) && (a.x!=0) && (b!=0) ){
            instability(&_ff_st_half_underflow_count);
        }
    }
    #endif
    
    return res;
}
#endif

#endif
