// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_NUMERICAL_NOISE
#define FF_ST_NUMERICAL_NOISE

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline int flexfloat_st<exp_bits, frac_bits>::numericalnoise() const
{
    int res = 1;
    double x0, x1, x2, xx;

    xx = x.getValue() + y.getValue() + z.getValue();
    if (xx == 0.0)
    {
        if ((x.getValue() == y.getValue()) && (x.getValue() == z.getValue()))
            res = 0;
    }
    else
    {
        xx = 3. / xx;
        x0 = x.getValue() * xx - 1.;
        x1 = y.getValue() * xx - 1.;
        x2 = z.getValue() * xx - 1.;
        res = x0 * x0 + x1 * x1 + x2 * x2 > 3.241001342318910E-02;
    }
    return res;
}

#endif