// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_STR_HPP
#define FF_ST_STR_HPP

#include <string.h>
#define MAXCHAINE 256
static char chstr[MAXCHAINE][64];
static int numstr = 0;

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline char *flexfloat_st<exp_bits, frac_bits>::str(char *s) const
{
    int nn, nres, naux;
    double fract_res, res;
    int tmp;
    char *t;

    // compute the number of exact significant digits always
    // with the same rounding mode
    fesetround(FE_TONEAREST);

    const uint_fast8_t MAX_PREC = (uint_fast8_t)((double)(frac_bits + 1) * LOG);

    if (accuracy == DIGIT_NOT_COMPUTED || accuracy == RELIABLE_RESULT)
        this->nb_significant_digit();

    if (accuracy == 0)
    {
        strcpy(s, " @.0");
    }
    else
    {
        accuracy = (accuracy < MAX_PREC) ? accuracy : MAX_PREC;
        // IDRIS/PC  : on evite l'"overflow"

        // res = (x.getValue()) / (double)(3) + (y.getValue()) / (double)(3) + (z.getValue()) / (double)(3);
        res = (x.getValue() + y.getValue() + z.getValue()) / 3.;
        if (res < 0.0)
            strcpy(s, "-0. ");
        else
            strcpy(s, " 0. ");
        res = fabs(res);

        if (res == 0.0)
        {
            nn = 0;
            fract_res = 0.0;
        }
        else
        {
            if (res >= 1.0)
                nn = (int)log10(res) + 1;
            else
                nn = (int)log10(res);
            fract_res = res * pow(10., -nn);
            if (fract_res < 0.1)
                fract_res = 0.1;
            if (fract_res >= 1.0)
            {
                fract_res = 0.1;
                nn = nn + 1;
            }
        }
        naux = accuracy + 3;
        t = s + 3;
        for (int i = 4; i < naux; i++)
        {
            nres = (int)(fract_res * 10.);
            *t++ = 48 + nres;
            fract_res = 10.0 * fract_res - nres;
        }
        tmp = (int)(fract_res * 10.0);
        nres = tmp < 9 ? tmp : 9;
        *t++ = 48 + nres;
        *t++ = 'E';
        if (nn < 0)
            *t++ = '-';
        else
            *t++ = '+';
        nn = abs(nn);
        *t++ = 48 + (nn / 100);
        nn = nn % 100;
        *t++ = 48 + (nn / 10);
        *t++ = 48 + (nn % 10);
        *t = '\0';
    }
    fesetround(FE_UPWARD);
    return (s);
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline char *strp(const flexfloat_st<exp_bits, frac_bits> &a)
{
    char *s;

    s = chstr[numstr];
    if ((++numstr) == MAXCHAINE)
        numstr = 0;
    return a.str(s);
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline void flexfloat_st<exp_bits, frac_bits>::display() const
{
    printf(" %+.16e -- %+.16e -- %+.16e \n", x.getValue(), y.getValue(), z.getValue());
}

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline std::ostream &operator<<(std::ostream &s, const flexfloat_st<exp_bits, frac_bits> &a)
{
    char ch[25];
    return s << a.str(ch);
}

#endif