// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FLEXFLOAT_ST_HPP
#define FLEXFLOAT_ST_HPP


//!\\ AJOUTER SI ON VEUT PRENDRE EN COMPTE LES DIFFERENTS COMPILATEUR //!\\ + METTRE COMMENTAIRE README

#ifdef FF_ST_FLOAT128
// #ifdef __GNUC__

#define FLEXFLOAT_FLOAT128

extern "C"{
  #include <quadmath.h>
}

// typedef __float128 float128;

// #elif defined(__INTEL_COMPILER)
// typedef _Quad float128;

// extern "C" {
// _Quad __ldexpq(_Quad, int);
// _Quad __frexpq(_Quad, int*);
// _Quad __fabsq(_Quad);
// _Quad __floorq(_Quad);
// _Quad __ceilq(_Quad);
// _Quad __sqrtq(_Quad);
// _Quad __truncq(_Quad);
// _Quad __expq(_Quad);
// _Quad __powq(_Quad, _Quad);
// _Quad __logq(_Quad);
// _Quad __log10q(_Quad);
// _Quad __sinq(_Quad);
// _Quad __cosq(_Quad);
// _Quad __tanq(_Quad);
// _Quad __asinq(_Quad);
// _Quad __acosq(_Quad);
// _Quad __atanq(_Quad);
// _Quad __sinhq(_Quad);
// _Quad __coshq(_Quad);
// _Quad __tanhq(_Quad);
// _Quad __fmodq(_Quad, _Quad);
// _Quad __atan2q(_Quad, _Quad);

// #define ldexpq __ldexpq
// #define frexpq __frexpq
// #define fabsq __fabsq
// #define floorq __floorq
// #define ceilq __ceilq
// #define sqrtq __sqrtq
// #define truncq __truncq
// #define expq __expq
// #define powq __powq
// #define logq __logq
// #define log10q __log10q
// #define sinq __sinq
// #define cosq __cosq
// #define tanq __tanq
// #define asinq __asinq
// #define acosq __acosq
// #define atanq __atanq
// #define sinhq __sinhq
// #define coshq __coshq
// #define tanhq __tanhq
// #define fmodq __fmodq
// #define atan2q __atan2q
// }

// inline _Quad isnanq(_Quad v)
// {
//    return v != v;
// }
// inline _Quad isinfq(_Quad v)
// {
//    return __fabsq(v) > 1.18973149535723176508575932662800702e4932Q;
// }
// #endif // compiler

#endif


#include <inttypes.h>
#include <iostream>
#include <cmath>

#include <flexfloat.hpp>
#include <flexfloat.h>

#include "random_st.h"


#ifdef FF_ST_HALF_EMULATION
#include <half.hpp>     // ils mettent <half.hpp> eux et #include <cadna_half.hpp> (voir add) // PAS SUR DE METTRE CES TAGS
using namespace half_float;
#endif


#define DIGIT_NOT_COMPUTED -1
#define RELIABLE_RESULT 255

#define FF_ST_DIV 0x1
#define FF_ST_MUL 0x2
#define FF_ST_POWER 0x4
#define FF_ST_MATH 0x8
#define FF_ST_BRANCHING 0x10
#define FF_ST_INTRINSIC 0x20
#define FF_ST_CANCEL 0x40
#define FF_ST_ALL 0xFFFFFFFF
#ifdef FF_ST_HALF_EMULATION
#define FF_ST_HALF_OVERFLOW    0x80
#define FF_ST_HALF_UNDERFLOW    0x100
#endif

extern unsigned int _ff_st_div_tag;
extern unsigned int _ff_st_mul_tag;
extern unsigned int _ff_st_power_tag;
extern unsigned int _ff_st_branching_tag;
extern unsigned int _ff_st_cancel_tag;
extern unsigned int _ff_st_intrinsic_tag;
extern unsigned int _ff_st_math_tag;
#ifdef FF_ST_HALF_EMULATION
extern unsigned int _ff_st_half_overflow_tag;
extern unsigned int _ff_st_half_underflow_tag;
#endif

extern unsigned int _ff_st_div_change;
extern unsigned int _ff_st_mul_change;
extern unsigned int _ff_st_power_change;
extern unsigned int _ff_st_branching_change;
extern unsigned int _ff_st_cancel_change;
extern unsigned int _ff_st_intrinsic_change;
extern unsigned int _ff_st_math_change;
#ifdef FF_ST_HALF_EMULATION
extern unsigned int _ff_st_half_overflow_change; 
extern unsigned int _ff_st_half_underflow_change; 
#endif

extern unsigned long _ff_st_div_count;
extern unsigned long _ff_st_mul_count;
extern unsigned long _ff_st_power_count;
extern unsigned long _ff_st_branching_count;
extern unsigned long _ff_st_cancel_count;
extern unsigned long _ff_st_intrinsic_count;
extern unsigned long _ff_st_math_count;
#ifdef FF_ST_HALF_EMULATION
extern unsigned long _ff_st_half_overflow_count; 
extern unsigned long _ff_st_half_underflow_count;
#endif 

extern const char *_ff_st_div_err_msg;
extern const char *_ff_st_mul_err_msg;
extern const char *_ff_st_power_err_msg;
extern const char *_ff_st_branching_err_msg;
extern const char *_ff_st_cancel_err_msg;
extern const char *_ff_st_intrinsic_err_msg;
extern const char *_ff_st_math_err_msg;
#ifdef FF_ST_HALF_EMULATION
extern const char *_ff_st_half_overflow_err_msg;
extern const char *_ff_st_half_underflow_err_msg;
#endif

extern unsigned int _ff_st_random;
extern int _ff_st_random_counter;
extern unsigned int _ff_st_recurrence;

extern int _ff_st_cancel_value;
extern int _ff_st_max_instability;
extern int _ff_st_instability_detected;

void instability(unsigned long *unstab);

void ff_st_init(int, unsigned int, unsigned int, unsigned int);
void ff_st_init(int, unsigned int);
void ff_st_init(int);
void ff_st_end();
// inline void  instability(unsigned long*) __attribute__((always_inline));

typedef union u_doublebf
{
    double fl;
    unsigned long bf;
} doublebf;

#ifdef FF_ST_FLOAT128
typedef union u_float128bf {
  __float128 fl;
  __uint128_t bf;
} float128bf;
#endif

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
class flexfloat_st
{
protected:
    flexfloat<exp_bits, frac_bits> x;
    flexfloat<exp_bits, frac_bits> y;
    flexfloat<exp_bits, frac_bits> z;
    mutable uint_fast8_t accuracy;

public:
    // Empty constructor --> initialize to positive zero.
    inline flexfloat_st()
    {
        // std::cout << "empty\n";
        x = flexfloat<exp_bits, frac_bits>();
        y = flexfloat<exp_bits, frac_bits>();
        z = flexfloat<exp_bits, frac_bits>();
        accuracy = DIGIT_NOT_COMPUTED;
    }

    inline flexfloat_st(fp_t a, fp_t b, fp_t c) : accuracy(-1)
    {
        // std::cout << "copy from fpt a b c\n";
        x = flexfloat<exp_bits, frac_bits>(a);
        y = flexfloat<exp_bits, frac_bits>(b);
        z = flexfloat<exp_bits, frac_bits>(c);
    }
    inline flexfloat_st(const fp_t &a) : accuracy(15)
    {
        x = flexfloat<exp_bits, frac_bits>(a);
        y = flexfloat<exp_bits, frac_bits>(a);
        z = flexfloat<exp_bits, frac_bits>(a);
    };

    inline flexfloat_st(const flexfloat_st &o)
    {
        // std::cout << "copy from flex ";
        x = o.x;
        y = o.y;
        z = o.z;
        accuracy = o.accuracy;
    }
    
    template <uint_fast8_t exp_bits2, uint_fast8_t frac_bits2>
    inline flexfloat_st &operator=(const flexfloat_st<exp_bits2,frac_bits2> &o)
    {
        // std::cout << "copy from temp\n";
        x = o.getx();
        y = o.gety();
        z = o.getz();
        accuracy = o.getaccuracy();
        return *this;
    }


    //         // Constructor from flexfloat types
    //     template <uint_fast8_t e, uint_fast8_t f> INLINE flexfloat (const flexfloat<e, f> &w) {
    //         v.value = w.getValue();
    //         v.desc.exp_bits = exp_bits;
    //         v.desc.frac_bits = frac_bits;
    //         flexfloat_sanitize(&v);
    //     }

    // Constructor from castable type
    // template <typename U>
    // INLINE flexfloat_st(const U &w)
    // {
    //     x = flexfloat<exp_bits, frac_bits>(w);
    //     y = flexfloat<exp_bits, frac_bits>(w);
    //     z = flexfloat<exp_bits, frac_bits>(w);
    //     accuracy = DIGIT_NOT_COMPUTED;
    // }

    flexfloat<exp_bits,frac_bits> getx() const { return (this->x); }
    flexfloat<exp_bits,frac_bits> gety() const { return (this->y); }
    flexfloat<exp_bits,frac_bits> getz() const { return (this->z); }
    uint_fast8_t getaccuracy() const { return (this->accuracy); }
    
    inline void ff_st_cast(flexfloat_st *obj, const flexfloat_st *source);
    inline float ff_st_get_float(const flexfloat_st *obj);
    inline double ff_st_get_double(const flexfloat_st *obj);
    inline long double ff_st_get_longdouble(const flexfloat_st *obj);

    inline int nb_significant_digit() const __attribute__((always_inline));
    inline int approx_digit() const __attribute__((always_inline));
    inline int numericalnoise() const __attribute__((always_inline));
    inline int computedzero() const __attribute__((always_inline));

#include "ff_st_intr.hpp"
#include "ff_st_convert.hpp"
#include "ff_st_math.hpp"

#include "ff_st_to.hpp"

#include "ff_st_add.hpp"
#include "ff_st_sub.hpp"
#include "ff_st_mul.hpp"
#include "ff_st_div.hpp"
#include "ff_st_fma.hpp"

#include "ff_st_eq.hpp"
#include "ff_st_ge.hpp"
#include "ff_st_gt.hpp"
#include "ff_st_le.hpp"
#include "ff_st_lt.hpp"
#include "ff_st_ne.hpp"

    char *str(char *) const;
    char *strp(const flexfloat_st<exp_bits, frac_bits> &);
    void display() const;

    // friend std::ostream& operator << (std::ostream& os, flexfloat_st& obj)
    // {
    //     os << "x : " << obj.x << "\n";
    //     os << "y : " << obj.y << "\n";
    //     os << "z : " << obj.z << "\n";
    //     os << "accuracy : " << obj.nb_significant_digit() << "\n";
    //     return os;
    // }
};

#include "ff_st_digitnumber.hpp"
#include "ff_st_str.hpp"
#include "ff_st_numerical_noise.hpp"
#include "ff_st_computedzero.hpp"
#include "ff_st_cast.hpp"

#endif
