// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#include <stdio.h>

#include "../include/random_st.h"

static unsigned int   recurrence = 987654321; 

// 0x40000 -> 262144
#define _ST_MAX_RANDOMTABLE 0x40000

unsigned short _st_random_table[_ST_MAX_RANDOMTABLE];
unsigned int   _st_table_counter=0;

static unsigned short _st_random_function()
{
	static const int IA=16087;
	static const int IM=2147483647;
	static const int IQ=127773;
	static const int IR=2836;
	static const int MASK=123459876;
	int k;
	
	recurrence^=MASK;
	k=recurrence/IQ;
	recurrence=IA*(recurrence-k*IQ)-IR*k;
	recurrence+=IM;
	recurrence^=MASK;
	
	// We need only a 16 bits number.
	// The lower and higher bits are not used.
	
	return (recurrence&0x00FFFF00)>>8;
}


void _st_init_table(unsigned int a)
{
	int i;
	recurrence+=a;
	for(i=0;i<_ST_MAX_RANDOMTABLE;i++){
		_st_random_table[i] = _st_random_function();
	}
}
